; Ranrot macros & data - SSE version
;
; These are high-quality high-speed uniform random number generators.
;
; Thanks to:
;  - Agner Fog for the original code, on which these macros are based (http://www.agner.org/random/)
;  - Mark Larson for his ideas on how to speed up the code
;

; Equates for the generator
Ranrot_SSE_JJ	equ	10	; Lag
Ranrot_SSE_KK	equ	17	; Circular buffer size
Ranrot_SSE_R1	equ	19	; Rotate count A
Ranrot_SSE_R2	equ	27	; Rotate count B

; Generates 256 bits of random data in XMM1:XMM0
;
; Uses: ebx, ecx, xmm0-xmm3
WBRandomSSE MACRO
	LOCAL c1, c2
	
	mov ebx, Ranrot_SSE_P1
	mov ecx, Ranrot_SSE_P2
	movapd xmm0, Ranrot_SSE_buf[ebx]
	movapd xmm1, Ranrot_SSE_buf[ebx+16]
	movapd xmm2, xmm0
	movapd xmm3, xmm1
	pslld xmm0, Ranrot_SSE_R1
	pslld xmm1, Ranrot_SSE_R2
	psrld xmm2, 32-Ranrot_SSE_R1
	psrld xmm3, 32-Ranrot_SSE_R2
	por xmm0, xmm2
	por xmm1, xmm3
	paddd xmm0, Ranrot_SSE_buf[ecx]
	paddd xmm1, Ranrot_SSE_buf[ecx+16]
	movapd Ranrot_SSE_buf[ebx], xmm1
	movapd Ranrot_SSE_buf[ebx+16], xmm0
	sub ebx, 16
	jnc c1
	mov ebx, (Ranrot_SSE_KK-1)*32
c1:
	sub ecx, 16
	jnc c2
	mov ecx, (Ranrot_SSE_KK-1)*32
c2:
	mov Ranrot_SSE_P1, ebx
	mov Ranrot_SSE_P2, ebx
ENDM

; Generates 128 bits of random data in XMM0
;
; Uses: ebx, ecx, xmm0-xmm3
WBRandomSSE2 MACRO
	LOCAL skip, calc
	
	cmp Ranrot_SSE_last, 0
	jz calc
	
	movapd xmm0, Ranrot_SSE_last
	mov Ranrot_SSE_last, 0
	jmp skip
calc:
	WBRandomSSE
	
	movapd Ranrot_SSE_last, xmm1
skip:
ENDM

; Generates two double-precision floating-point numbers and stores them in XMM0
;
; Uses: ebx, ecx, xmm0-xmm3
WDRandomSSE MACRO

	WBRandomSSE2
	
	movapd xmm1, Ranrot_SSE_double_mask
	movapd xmm2, Ranrot_SSE_double_set
	movapd xmm3, Ranrot_SSE_double_one
	
	pand xmm0, xmm1
	por xmm0, xmm2
	subpd xmm0, xmm3

ENDM

; Generates four single-precision floating-point numbers and stores them in XMM0
;
; Uses: ebx, ecx, xmm0-xmm3
WSRandomSSE MACRO

	WBRandomSSE2
	
	movapd xmm1, Ranrot_SSE_single_mask
	movapd xmm2, Ranrot_SSE_single_set
	movapd xmm3, Ranrot_SSE_single_one
	
	pand xmm0, xmm1
	por xmm0, xmm2
	subps xmm0, xmm3

ENDM

; Initializes the generator
;
; Uses: eax, ebx, ecx, xmm0-xmm3
WRandomInitSSE MACRO seed:req
	LOCAL _loop, _loop2

	mov eax, seed
	xor ecx, ecx
_loop:
	imul eax, 2891336453
	add eax, 1
	mov Ranrot_SSE_buf[ecx*4], eax
	add ecx, 1
	cmp ecx, Ranrot_SSE_KK*8
	jb _loop
	
	mov Ranrot_SSE_P1, 0
	mov Ranrot_SSE_P2, Ranrot_SSE_JJ*32
	
	WBRandomSSE
	
	push edi
	mov edi, 30
_loop2:
	WBRandomSSE
	sub edi, 1
	jnz _loop2
	
	pop edi
ENDM

; Some data for the generators
.data

align 16
Ranrot_SSE_P1		DD	0							; First pointer in circular buffer
Ranrot_SSE_P2		DD	0							; Second pointer in circular buffer

align 16
Ranrot_SSE_buf	DD	(8*Ranrot_SSE_KK) dup (0); Circular buffer

align 16
Ranrot_SSE_double_one		REAL8	1.0, 1.0
Ranrot_SSE_double_mask	DD 0FFFFFFFFh, 000FFFFFh, 0FFFFFFFFh, 000FFFFFh
Ranrot_SSE_double_set		DD 0, 3FF00000h, 0, 3FF00000h

Ranrot_SSE_single_one		REAL4	1.0, 1.0, 1.0, 1.0
Ranrot_SSE_single_mask	DD 007FFFFFh, 007FFFFFh, 007FFFFFh, 007FFFFFh
Ranrot_SSE_single_set		DD 3F800000h, 3F800000h, 3F800000h, 3F800000h

Ranrot_SSE_last	DD 0,0,0,0
